#lang racket

;; 12-7.  Here is our final version.  We'll clean up all the
;; intermediate comments.  We'll also have a world with TWO boxes,
;; each with its own bouncing objects.  A keystroke will add the shape
;; to BOTH boxes.

;; we'll do a little cleanup while we're at it.

;; start this with (run).

(require rackunit)
(require "extras.rkt")
(require 2htdp/universe)
(require 2htdp/image)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; INTERFACES
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Stateful Objects:  they don't return a new WorldObj-- they just
;; change their state
(define StatefulWorldObj<%>
  (interface ()
    on-tick       ; ->  Void
    on-mouse      ; Integer Integer MouseEvt -> Void
    on-key        ; KeyEvt -> Void
    add-to-scene  ; Scene -> Scene
    ))

(define StatefulContainer<%>
  (interface ()
    add-object  ; WorldObj<%> -> Void
    add-stateful-object ; StatefulWorldObj<%> -> Void
    ))

;; functional WorldObj's.
(define WorldObj<%>
  (interface ()
    on-tick       ; -> WorldObj<%>
    on-mouse      ; Integer Integer MouseEvt -> WorldObj<%>
    on-key        ; KeyEvt -> world-obj
    add-to-scene  ; Scene -> Scene
    ))

(define Subscriber<%>
  (interface ()
    ;; Integer -> Void
    ;; EFFECT: record new right edge
    change-edge  
    ))

(define Publisher<%>
  (interface ()
    ;; Subscriber<%> -> Void
    ;; EFFECT: add a new subscriber
    subscribe  
    ))

;; the extra operations supplied by a Box
(define BoxExtras<%>
  (interface ()
    
    ;; the coordinates of the center
    get-x ; -> Integer
    get-y ; -> Integer
    
    ;; the x-position of the two edges
    left-edge ;  -> Integer
    right-edge ; -> Integer
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; CLASSES
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;
;;; World%
;;;;;;;;;;;;;;;;

;; A World% contains a list of WorldObj<%>s and StatefulWorldObj<%>s
;; All it does is to distribute each of the world-obj methods to each
;; of its objects.
;; A World is a (new World% 
;;                     [objects ListOf<WorldObj<%>>]
;;                     [stateful-objects 
;;                       ListOf<StatefulWorldObj<%>>
;;                       default: empty])
(define World%
  (class* object% (StatefulWorldObj<%> StatefulContainer<%>)
    (init-field objects)                  ;; ListOf<WorldObj<%>>
    (init-field [stateful-objects empty]) ;; ListOf<StatefulWorldObj<%>>
     
    (super-new)
    
    ;; WorldObj<%> -> Void
    ;; EFFECT: Add the given object to the objects field
    (define/public (add-object obj)
      (set! objects
            (cons obj objects)))

    ;; StatefulWorldObj<%> -> Void
    ;; EFFECT: Add the given stateful-object to the stateful-objects
    ;; field. 
    (define/public (add-stateful-object obj)
      (set! stateful-objects (cons obj stateful-objects)))

    ;; -> Void
    ;; EFFECT: pass on-tick to each of the functional objects, and
    ;; collect the result in the objects field; then perform the
    ;; effect on each of the stateful objects.
    (define/public (on-tick)
      (set! objects
            (map 
             (lambda (obj) (send obj on-tick))
             objects))
      (for-each
       (lambda (obj) (send obj on-tick))
       stateful-objects))

    ;; Integer Integer MouseEvt -> Void
    ;; We pass each mouse event to every object in the world
    ;; and let that object decide whether to respond to it.
    ;; Change is similar to on-tick
    (define/public (on-mouse x y mev)
      (set! objects
            (map
             (lambda (obj) (send obj on-mouse x y mev))
             objects))
      (for-each
       (lambda (obj) (send obj on-mouse x y mev))
       stateful-objects))
       
    
    ;; KeyEvent -> Void
    (define/public (on-key kev)
      (set! objects
            (map
             (lambda (obj) (send obj on-key kev))
             objects))
      (for-each
       (lambda (obj) (send obj on-key kev))
       stateful-objects))

    ;; note the "map" above.  That means that if you send a message to
    ;; a functional object, and the functional object adds an object
    ;; to the world, that new object will be thrown away!  [This
    ;; is hard-won knowledge, which I had forgotten and had to
    ;; painfully reconstruct.  Wed Nov 09 15:53:45 2011]
    
    ;; do you see repeated code here?  
    ;; Is there an opportunity for an abstraction?
    
    ; Scene -> Scene
    ; implementation: first paint the stateful objects, then the
    ; functional ones. 
    (define/public (add-to-scene s)
      (local
       ((define s2 (foldr
                    (lambda (obj s1) (send obj add-to-scene s1))
                    s
                    stateful-objects)))
      (foldr
       (lambda (obj s1) (send obj add-to-scene s1))
       s2
       objects)))

    ;; below this line are getters introduced for testing only

    ;; -> (list ListOf<<WorldObj<%>> ListOf<StatefulWorldObj<%>>)
    (define/public (for-test:get-all-objects)
      (list objects stateful-objects))

    ))   

;;;;;;;;;;;;;;;;
;;; Box%
;;;;;;;;;;;;;;;;


;; A Box% represents a  resizeable rectangle
;; It will resize itself in response to mouse drags.
;; A Box is a (new Box% [x Integer][y Integer]
;;                      [w Integer][h Integer])
(define Box%
  (class* object% (StatefulWorldObj<%> Publisher<%> BoxExtras<%>)
    (init-field x)
    (init-field y)
    (init-field w)
    (init-field h)
    ;; x y : Integers: position of center
    ;; w h : Integers: width and height, in pixels
    (field [selected? false])

    ;;;;;;;;;;;;;;;;
    ;; implement Publisher<%> interface
    ;;;;;;;;;;;;;;;;

    (field [subscribers empty])         ; ListOf<Subscriber<%>>

    ;; Subscriber<%> -> Void
    ;; EFFECT: Add the subscriber to the subscriber list
    (define/public (subscribe subscriber)
      (set! subscribers (cons subscriber subscribers)))
    
    ;; Integer -> Void
    ;; EFFECT: publish the value to all the subscribers
    (define (publish val)
      (for-each
        (lambda (obj)
          (send obj change-edge val))
        subscribers))

    (super-new)
    

    ;; Implement the BoxExtras<%> interface.  The factory relies on these.

    (define/public (left-edge) (- x (/ w 2)))
    (define/public (right-edge) (+ x (/ w 2)))

    (define/public (get-x) x)
    (define/public (get-y) y)
    
    ;; notice nobody gets to see its height

    ;; doesn't respond to ticks or key events
    (define/public (on-tick) this)
    (define/public (on-key kev) this)

    ;; on-mouse: Integer Integer MouseEvent -> Void
    ;; EFFECT:
    ;; mouse-down near right edge of box => mark box selected
    ;; drag & selected => make right edge follow mouse
    ;; mouse-up => mark box unselected

    ;; if the mouse is dragging "near" the right edge of this box,
    ;; resize the box to match the given mouse posn, otherwise return this
    ;; box unchanged.  
    ;; if we move the right edge, we have to adjust the center to match
    (define/public (on-mouse mouse-x mouse-y evt)
      (cond
        [(mouse=? evt "button-down")
         (send this after-button-down mouse-x mouse-y)]
        [(mouse=? evt "drag")
         (send this after-drag mouse-x)]
        [(mouse=? evt "button-up")
         (send this unselect)]))


    ;; EFFECT: if the button-down is near the right edge, mark this
    ;; box as selected
    (define/public (after-button-down mouse-x mouse-y)
      (if (near-right-edge? mouse-x mouse-y)
        (set! selected? true)
        this))
      
    ;; EFFECT: if this box is selected, move the right edge to the
    ;; mouse position.
    (define/public (after-drag mouse-x)
      (if selected?
        (local
          ((define new-width (- mouse-x (left-edge))))
          (adjust-width new-width))
        this))
 
    ;; EFFECT: unselect the box.
    (define/public (unselect)
      (set! selected? false))

    ;; adjust-width : Integer -> Void
    ;; GIVEN: A value for the new width of this box
    ;; EFFECT: adjust the center and width so that the width of this
    ;; box is new-width and the left edge is unchanged.
    ;; NEW EFFECT: Publish the new right edge to the subscribers
    ;; ALGORITHM: To do this, we have to restore the invariant that 
    ;; right-edge = x + w/2
    ;; left-edge = x - w/2
    ;; left-edge isn't supposed to change, so we adjust x accordingly
    ;; the objects stay in the same place on the screen, so they stay
    ;; the same.
    (define (adjust-width new-width)
      (set! x (+ (send this left-edge) (/ new-width 2)))
      (set! w new-width)
      (publish (+ x (/ w 2))))

    ;; GIVEN: the coordinates of a point
    ;; RETURS: true iff the point is "near" the right edge of this
    ;; box. 
    (define (near-right-edge? other-x other-y)
      (near-vertical-line? 
        (send this right-edge)         ; the x-coordinate of the
                                        ; vertical line
        (- y (/ h 2))                  ; the y-coords of the endpoints
        (+ y (/ h 2))    
        other-x other-y))

    (define NEAR-VERTICAL-LINE-THRESHOLD 5)

    ;; Integer^5 -> Boolean
    ;; returns true iff other-pos is "near" the line from posn1 to posn2
    (define (near-vertical-line? line-x y1 y2 other-x other-y)
      (and
        (<= y1 other-y)
        (<= other-y y2)
        (<= (abs (- line-x other-x)) NEAR-VERTICAL-LINE-THRESHOLD)))

    (define/public (add-to-scene s)
      (place-image 
        (rectangle w h "outline" "black")
        x y s))

    ))


(begin-for-test

  (local
    ((define box1 (new Box% [x 200][y 50][w 100][h 20])))
    (check-equal? (send box1 left-edge) 150)
    (check-equal? (send box1 right-edge) 250)
    (send box1 on-mouse 252 50 "button-down")
    (send box1 on-mouse 254 50 "drag")  
    (check-equal? (send box1 left-edge) 150)
    (check-equal? (send box1 right-edge) 254)) 
 
   )

;;;;;;;;;;;;;;;;
;;; DraggableObject%
;;;;;;;;;;;;;;;;

;; A DraggableObject% represents a draggable object
;; A DraggableObject is a (new DraggableObject% 
;;                          [x Integer][y Integer]
;;                          [box Box<%>] 
;;                          [speed PosInteger])

(define DraggableObject%
  (class* object% (StatefulWorldObj<%> Subscriber<%>)
    (init-field
     x y        ; the position of the center of the ball
     box        ; the Box% in which the ball lives
     speed)     ; ball speed in pixels/tick (either positive or
                ; negative)

    ;; we expected these to be initialized in a super-new
    (init-field selected-image)
    (init-field unselected-image)

    ;; fields for selection and dragging
    (field [selected? false])      
    ;; is the object selected?. Initially false.
    (field [mx false][my false])              
    ;; MaybeInt: if selected, the coordinates of the mouse RELATIVE TO
    ;; THE CENTER, otherwise false.

    ;; the left-edge and right-edge fields.
    ;; INVARIANT: left-edge and right-edge are the box's left and right
    ;; edges. 

    ;; initialize right edge from the box, after that it will get
    ;; updated by messages.
    (field [right-edge (send box right-edge)])
    ;; the left edge doesn't change
    (field [left-edge (send box left-edge)])

    (super-new)

    ;; at initialization, subscribe to the box's postings
    (send box subscribe this)

    ;; implement Subscriber<%> interface.
    ;; Integer -> Void
    ;; EFFECT: use the received message to update right-edge
    (define/public (change-edge n)
      (set! right-edge n))

    ;; add-to-scene : Scene -> Scene
    ;; RETURNS: a scene like the given one, but with this object painted on it.
    (define/public (add-to-scene s)
      (place-image
        ;; (send this get-image)
        (if selected? 
          selected-image
          unselected-image)
        x y s))

    ;; on-mouse : Integer Integer MouseEvent -> Void
    ;; EFFECT: act on the mouse event
    (define/public (on-mouse nx ny mev)
      (cond
        [(mouse=? mev "button-down")
         (send this after-button-down nx ny)]
        [(mouse=? mev "drag")
         (send this after-drag nx ny)]
        [(mouse=? mev "button-up")
         (send this after-button-up)]
        [else this]))

    ;; EFFECT: if the button-down is inside the object, then
    ;; set selected? to true and set mx, my to mouse position relative
    ;; to x,y.
    (define/public (after-button-down nx ny)
      (if (send this inside-this? nx ny)
        (begin
          (set! selected? true)
          (set! mx (- nx x))
          (set! my (- ny y)))
        this))

    ;; EFFECT: if the object is selected, then move it to correct
    ;; relative position.
    (define/public (after-drag nx ny)
      (if selected?
        (begin
          (set! x (- nx mx))
          (set! y (- ny my)))  ;  selected is already true
        this))

    ;; EFFECT: unselect this object
    (define/public (after-button-up)      
      (set! selected? false)
      (set! mx false)
      (set! my false))

    ;; don't respond to  key events
    (define/public (on-key kev) this)

    ; -> Void%
    ; EFFECT: sets the object to the state it should be in after a tick.
    (define/public (on-tick)
      (cond
        (selected? this)
        ((send this would-hit-right-edge?)
         (send this place-at-right-edge))
        ((send this would-hit-left-edge?)
         (send this place-at-left-edge))
        (else 
          (set! x (+ x speed)))))

    ;; getters for testing only -- also common to Ball% and Square%:
    (define/public (for-test:get-x) x)
    (define/public (for-test:get-y) y)
    (define/public (for-test:get-speed) speed)   

    ))

;; Here are the three shapes.  Notice that most of the functionality
;; is in the superclass.  We need only a little shape-specific stuff
;; for each of them.

(define Ball%
  (class* DraggableObject%
    (StatefulWorldObj<%> Subscriber<%>)

    (inherit-field
     x y        ; the position of the center of the ball
     box        ; the Box% in which the ball lives
     speed      ; ball speed in pixels/tick (either positive or negative)
     )

    ;; also inherit these from the superclass
    (inherit-field selected? left-edge right-edge)

    ;; specific to ball
    (field [radius 20])   

    ;; initialize the superclass, passing the images for the selected
    ;; and unselected cases.
    (super-new
      [selected-image (circle radius "solid" "red")]
      [unselected-image   (circle radius "outline" "red")])

    ;; the ball-specific geometry methods, cleaned up

    (define/public (place-at-right-edge)
        (set! x (- right-edge radius))
        (set! speed (- speed)))

    (define/public (place-at-left-edge)
        (set! x (+ left-edge radius))
        (set! speed (- speed)))

    (define/public (would-hit-right-edge?)
      (>= (+ (+ x radius) speed) right-edge))
    
    (define/public (would-hit-left-edge?)
      (<= (+ (- x radius) speed) left-edge))
    
    (define/public (inside-this? mouse-x mouse-y)
      (<= 
       (+
        (sqr (- x mouse-x))
        (sqr (- y mouse-y)))
       (sqr radius)))
    
    ))

;; FlashingBall% is like a Ball%, but it displays differently: it
;; changes color on every other redisplay
(define FlashingBall%
  (class* 
    Ball%                               ; inherits from Ball%
    (StatefulWorldObj<%>)

    (field [color-change-interval 4])   ; how much time between color changes?
    (field [time-left color-change-interval])  ; how much time left
                                        ; til next color change

    (field [colors (list "red" "green")])  ; the list of possible
                                        ; colors, first elt is current color
    
    ;; here are fields of the superclass that we need.
    ;; note that we have to make radius a field rather than a constant.
    (inherit-field radius x y selected?)   

    (super-new)

    ;; Scene -> Scene
    ;; RETURNS: a scene like the given one, but with the flashing ball
    ;; painted on it.
    ;; EFFECT: decrements time-left and changes colors if necessary
    (define/override (add-to-scene s)
      (begin
        ;; is it time to change colors?
        (if (zero? time-left)
          (change-colors)
          (set! time-left (- time-left 1)))
        ;; now paint yourself on the scene
        (place-image
          (circle radius
            (if selected? "solid" "outline")
            (first colors))
          x y s)))

    ;; -> Void
    ;; EFFECT: rotate the list of colors, and reset time-left
    (define (change-colors)
      (set! colors (append (rest colors) (list (first colors))))
      (set! time-left color-change-interval))
    
    ))

;; Square% represents a square.
;; it is just like Ball, except:
;; -- it is represented by a green outline square of size 40
;; -- it has different geometry, so the bounce methods will be
;;    different. 
(define Square%
  (class* DraggableObject%
    (StatefulWorldObj<%> Subscriber<%>)
    (inherit-field
     x y        ; the position of the center of the square
     box        ; the Box% in which the square lives
     speed      ; speed in pixels/tick (either positive or negative)
     )

    ;; instead, inherit these from the superclass
    (inherit-field selected? left-edge right-edge)

    (field [size 40])

    (super-new
      [selected-image (square size "solid" "purple")]
      [unselected-image (square size "outline" "green")])
    
    ;; square-specific geometry methods:

    (define/public (place-at-right-edge)
      (set! x (- right-edge (/ size 2)))
      (set! speed (- speed)))

    (define/public (place-at-left-edge)
      (set! x (+ left-edge (/ size 2)))
      (set! speed (- speed)))

    (define/public (would-hit-right-edge?)
      (>= (+ (+ x (/ size 2)) speed) right-edge))
    
    (define/public (would-hit-left-edge?)
      (<= (+ (- x (/ size 2)) speed) left-edge))

    (define/public (inside-this? mouse-x mouse-y)
      (and
       (<= (- x (/ size 2)) mouse-x (+ x (/ size 2)))
       (<= (- y (/ size 2)) mouse-y (+ y (/ size 2)))))

    ))

;; BallFactory% 
;; implements StatefulWorldObj<%>
;; on key event "n", adds a new Ball at the center of the given box in
;; the given world; on "f" adds new FlashingBall
;; WHERE: the box must be a stateful-object in the given
;; world.
(define BallFactory%
  (class* object% (StatefulWorldObj<%>)

   (init-field world)            ; the World%
   (init-field box)              ; the Box%
   (init-field [speed 3])        ; the speed of the balls to be
                                 ; created.
                                        ; we make this different from
                                        ; the initial ball.
   (super-new)

   ;; ignore on-tick, on-mouse
   (define/public (on-tick) this)
   (define/public (on-mouse x y mev) this)

   ;; don't add yourself to the scene
   (define/public (add-to-scene s) s)

   ;; KeyEvent -> Void
   ;; catch "n" events here and create a new ball. 
    (define/public (on-key kev)
      (cond
        [(key=? kev "n") 
         (send this add-object Ball%)]
        [(key=? kev "f") 
         (send this add-object FlashingBall%)]
        [(key=? kev "s")
         (send this add-object Square%)]))
         ;; in #lang racket, the else is optional.

    ;; Class -> Void
    ;; EFFECT: Adds a new object of the given class at the center of
    ;; the box, and adds it to the container as a stateful object.
    ;; INVARIANT: the class must be initializable as
    ;; (new class [x Number][y Number][box Box][speed Number])
    ;; NOTE: this is a 
    (define/public (add-object class)
      (local
       ((define the-object
          (new class
               [x (send box get-x)]
               [y (send box get-y)]
               [box box]
               [speed speed])))
       (send world add-stateful-object the-object)))

    ))


;; -> World%
;; the initial world consists of a box and a ball
;; the ball is stateful, so it goes in "stateful-objects".
;; the box is stateful, so it goes in "stateful-objects".
;; V8: we make this a function of no arguments, so (run) will get a
;; fresh world every time.
#;(define (init-world)
  (local
   ((define the-box  (new Box% [x 100][y 45][w 150][h 75][selected? false]))
    (define the-ball (new Ball% [x 100][y 45][box the-box][speed 5]))
    (define the-world
      (new World%
           [objects          empty]
           [stateful-objects (list the-ball the-box)])))
   (send the-world
         add-stateful-object
         (new BallFactory%
              [world the-world]
              [box   the-box]))
   the-world))

;; -> World%
;; The initial world consists of two boxes, each containing a factory,
;; but no balls (or squares or flashing balls)
(define (init-world)
  (local
    ((define box1
       (new Box% [x 100][y 50][w 150][h 75]))
     (define box2
       (new Box% [x 100][y 450][w 150][h 75]))
     (define the-world
       (new World%
         [objects   empty]
         [stateful-objects (list box1 box2)])))
    (send the-world
      add-stateful-object
      (new BallFactory%
        [world the-world]
        [box   box1]))
    (send the-world
      add-stateful-object
      (new BallFactory%
        [world the-world]
        [box   box2]))
    the-world))

;; main function (run-world).  Create a new world and run it.
;; since on-tick, etc, return Void, we need to explicitly pass back w
;; to big-bang.
;; World% -> World%
(define (run-world initial-world)
  (big-bang 
   initial-world
   (on-tick
    (lambda (w) (send w on-tick) w))
   (on-mouse
    (lambda (w x y evt) (send w on-mouse x y evt) w))
   (on-draw
    (lambda (w) (send w add-to-scene (empty-scene 400 800))))
   (on-key
    (lambda (w kev) (send w on-key kev) w))))

;; -> World%
(define (run) (run-world (init-world)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; TESTING
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Create Abstractions to eliminate repeated code in testing

;;  Box<%> -> (list Integer Integer)
(define (get-edges box) (list (send box left-edge) (send box right-edge)))

;; World% (list Integer Integer) [String] -> Check.
;; RETURNS: a check that checks whether the object and stateful-object
;; lists in the given world have the given lengths
;; str is optional error message, default is empty string
(define (check-lengths w lst [str ""])
  (check-equal?
   (map length (send w for-test:get-all-objects))
   lst
   str))

(begin-for-test  

  ;; check dragging of wall

  (local
    ;; create a box and check its edges
    ((define box1 (new Box% [x 200][y 50][w 100][h 20])))
    (check-equal? (get-edges box1) 
      (list 150 250) 
      "edges should be initialized correctly")
    ;; send the box a button-down
    (send box1 on-mouse 252 50 "button-down")
    ;; send the box a drag message
    (send box1 on-mouse 253 50 "drag")
    ;; now the edges of the box should have changed
    (check-equal? (get-edges box1) 
      (list 150 253) 
      "left edge should stay the same, right edge should change"))

  ;; check simple ball motion"
 
  (local
    ;; first create the objects
    ((define the-box  (new Box% [x 100][y 45][w 100][h 75]))
     (define the-ball (new Ball% [x 100][y 45][box the-box][speed 5])))
    (send the-ball on-tick)
    (check-equal?
     (send the-ball for-test:get-x)
     (+ 100 5) 
     "ball should have moved 5 pixels")
    ;; see how the ball responds to mouse events
    (send the-ball on-mouse 105 45 "button-down")
    (check-equal? 
      (send the-ball for-test:get-speed)
      5
      "button down should not change speed")
    (send the-ball on-mouse 85 100 "drag")
    (check-equal? 
      (send the-ball for-test:get-speed)
      5
      "drag should  not change speed")
    (send the-ball on-mouse 85 100 "button-up")
    (check-equal? 
      (send the-ball for-test:get-speed)
      5
      "button-up should not change speed"))
  
  ;; check ball bouncing off wall

  (local
    ;; first create the objects
    ((define the-box  (new Box% [x 100][y 45][w 100][h 75]))
     ;; right edge of box is at 150.  So put the ball close to the edge.
     ;; center at edge - speed - radius + 1, so ball should bounce on next tick
     ;; x = 150 - 10 - 15 + 1 = 126
     (define the-ball
       (new Ball% [x 126][y 45][box the-box][speed 10])))
     ;(define ball-after-tick  (send the-ball on-tick)))
    (send the-ball on-tick)
    ;; check to see that the speed is now -10
    (check-equal? 
      (send the-ball for-test:get-speed)
      -10
      "after bounce, ball speed should be -10"))
        
  ;; check that ball doesn't bounce when wall is dragged
  
  (local
    ;; first create the objects
    ((define the-box  (new Box% [x 100][y 45][w 100][h 75])))
    ;; right edge of box is at 150.  So put the ball close to the edge.
    ;; center at edge - speed - radius + 1, so ball should bounce on next tick
    ;; x = 150 - 10 - 15 + 1 = 126
    (define the-ball (new Ball% [x 126][y 45][box the-box][speed 10]))
    ;; select the box
    (send the-box on-mouse 152 50 "button-down")
    ;; drag the box
    (send the-box on-mouse 165 50 "drag")
    (check-equal?
      (send the-box right-edge)
      165
      "right edge of box should have moved to mouse location 165")
    (check-equal?
     (send the-ball for-test:get-speed)
     10)
    (check-equal? (send the-ball for-test:get-x) 126)
    (check-equal? (send the-ball for-test:get-speed) 10)
      ;; now ball shouldn't be bouncing
      (send the-ball on-tick)
    (check-equal? (send the-ball for-test:get-x) 136)
    ;; check to see that the speed is still 10
      (check-equal? 
        (send the-ball for-test:get-speed)
        10
        "after wall drag, ball should still be moving right"))




)
  
 
