#lang racket

(require "interfaces.rkt")
(require 2htdp/universe)
(require 2htdp/image)

(provide Display%)

;; Display% implements a digital display of the Timer%.

;; Display% is smoothly-draggable, like Slider%
;; responds to keyboard + and - by incrementing or decrementing the
;; timer by 5.

(define Display%
  (class* object%
    (Controller<%>)

    (init-field x y)  ; x y position of center of the display

    (init-field timer) ; the timer to which this controller is connected

    (field [timer-value false])  ; Maybe<Number> .  The current value of the
                          ; timer if known

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    (define SIZE 20)

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    ;;; INITIALIZATION

    (super-new)
    ;; subscribe to updates from the timer
    (send timer subscribe (lambda (n) (set! timer-value n)))

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    ;;; ADD-TO-SCENE
    
    (define/public (add-to-scene s)
      (place-image
        (text (value->text timer-value) SIZE "green")
        x y s))

    ;; Maybe<Number> -> String
    (define (value->text v)
      (if (false? v) "??"
        (number->string v)))

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    ;;; ON-MOUSE

    ;;;  need the following additional state:
    (field [selected? false])
    ;; keep track of the last mouse event of any kind
    ;; ie, the "current" mouse location.
    ;; this will be set on mouse-enter.
    (field [mouse-x 0])
    (field [mouse-y 0])
    ;; position of button-down when the display was last selected.
    ;; recorded relative to (x,y)
    (field [last-button-down-x 0])
    (field [last-button-down-y 0])
    
    ;; Number Number MouseEvent -> Void
    ;; SD on mev
    (define/public (on-mouse mx my mev)
      (record-current-mouse-location mx my)
      (cond
        [(mouse=? mev "button-down") (on-button-down mx my)]
        [(mouse=? mev "drag")        (on-drag mx my)]
        [(mouse=? mev "button-up")   (on-button-up)]))

    ;; on-button-down : 
    (define (on-button-down mx my)
      (cond
        [(near-this? mx my) (select-this mx my)]
        [else false]))

    ;; on-drag : if this display is selected, then drag it smoothly,
    ;; otherwise ignore.
    (define (on-drag mx my)
      (cond
        [selected? (drag-this mx my)]))

    ;; on-button-up : this display is selected, just unselect
    (define (on-button-up)
      (cond
        [selected? (unselect-this)]))

    (define (near-this? mx my)
      (near-rectangle? mx my x y (* 3 SIZE) SIZE))

    ;; record the position of the button-down
    (define (select-this mx my)
      (set! last-button-down-x (- mx x))
      (set! last-button-down-y (- my y))
      (set! selected? true))

    ;; move center of slider to same position relative to the mouse.
    (define (drag-this mx my)
      (set! x (- mx last-button-down-x))
      (set! y (- my last-button-down-y)))

    (define (unselect-this)
      (set! selected? false))

    (define (record-current-mouse-location mx my)
      (set! mouse-x mx)
      (set! mouse-y my))

    ;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    ;; this works right with 1 display showing, but not with 2.
    ;; why??-- because BOTH displays are sending commands!
    ;; so need to keep track of mouse position!

    ;; check to see if current mouse location is near this display.
    ;; if so, do something
    (define/public (on-key kev) 
      (if (near-this? mouse-x mouse-y)
        (cond
          [(key=? kev "+") (send timer set-time! (+ timer-value 5))]
          [(key=? kev "-") (send timer set-time! (- timer-value 5))])
        false))
      
      
    ))

;; this should be moved to a single location and required...

(define TOLERANCE 5)

;; a point is near a rectangle if it's within TOLERANCE in both x and
;; y.
;; given the coordinates of the point px and py, and the rectangle at
;; center (x,y) and dimensions (w,h)
(define (near-rectangle? px py x y w h)
  (and
    (<= (- x (/ w 2) TOLERANCE) px (+ x (/ w 2) TOLERANCE))
    (<= (- y (/ h 2) TOLERANCE) py (+ y (/ h 2) TOLERANCE))))

