#lang racket

(require "interfaces.rkt")
(require 2htdp/universe)
(require 2htdp/image)

(provide Slider%)

;; Slider% implements a slider, which is attached to a Timer%.

;; This is the most complicated class in the system, because it has a
;; bunch of behaviors.

;; subscribes to the timer to get the timer current value.

;; display:
;; -- a horizontal line (the range) and a vertically-oriented
;; rectangle (the handle).
;; -- also displays the value of the timer, as a number, displayed
;; over the handle.

;; mouse response:

;; the handle can be selected for dragging in the usual way.
;; -- drag in x direction only.
;; -- it should use "smooth drag" (keep track of mouse-x relative to
;;    handle-x) but it doesn't (sorry)
;; -- on deselect, send a set-time! command to the timer.

;; the range can be selected for dragging 
;; -- only if the handle is not selected.
;; -- drag in 2 dimensions
;; -- uses smooth drag.

;; a Slider% is a (new Slider% [x Integer][y Integer][timer Timer%])
(define Slider%
  (class* object%
    (Controller<%>)

    (init-field x y)  ; x y position of center of the slider

    (init-field timer) ; the timer to which this slider is connected

    (field [timer-value false])  ; Maybe<Integer> .  The current value of the
                          ; timer if known

    (field [handle-x-position x])  ;; screen x position of the handle

    (field [handle-selected? false])
    (field [range-selected? false])
    ;; INVARIANT: at most one of these is true at any time.

    (define RANGE-WIDTH 150)
    (define RANGE-HEIGHT 2)
    (define HANDLE-WIDTH 10)
    (define HANDLE-HEIGHT 20)

    (define HALF-LENGTH (/ RANGE-WIDTH 2))

    (define RANGE-IMAGE 
      (beside
        (line 0 10 "black")
        (rectangle RANGE-WIDTH RANGE-HEIGHT "solid" "black")
        (line 0 10 "black")))

    (define HANDLE-IMAGE (rectangle HANDLE-WIDTH HANDLE-HEIGHT "solid"
                           "red")) 

    (define MAX-TIMER-VALUE 100)

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    ;;; INITIALIZATION

    (super-new)
    ;; subscribe to updates from the timer
    (send timer subscribe (lambda (n) (receive-timer-update n)))

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    ;; update timer-value and handle-x-position
    (define (receive-timer-update v)
      (set! timer-value v)
      (set! handle-x-position (value-to-handle-x v)))

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    ;;; ADD-TO-SCENE
    
    ;; first draw the range, then the handle, then the value
    (define/public (add-to-scene s)
      (place-value-image handle-x-position y timer-value
        (place-image HANDLE-IMAGE handle-x-position y
          (place-image RANGE-IMAGE x y s))))

    ;; in initial version, we didn't render the value
    ; (define (place-value-image x y v s) s)

    (define (place-value-image x y v s) 
      (place-image
        (text (value->text v) 16 "green")
        x (- y 20)
        s))

    ;; Maybe<Integer> -> String
    (define (value->text v)
      (if (false? v) "??"
        (number->string v)))

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    ;;; ON-MOUSE

    ;; Integer Integer MouseEvent -> Void
    ;; SD on mev
    (define/public (on-mouse mx my mev)
      (cond
        [(mouse=? mev "button-down") (on-button-down mx my)]
        [(mouse=? mev "drag")        (on-drag mx my)]
        [(mouse=? mev "button-up")   (on-button-up)]))

    ;; on-button-down : select the handle first, otherwise the range.
    (define (on-button-down mx my)
      (cond
        [(and (near-handle? mx my) 
           (not range-selected?))
         (select-handle)]
        [(and (near-range? mx my)
           (not handle-selected?))
         (select-range mx my)]
        [else false]))

    ;; on-drag : if handle is selected, move it, if range is selected,
    ;; move it, otherwise ignore.
    ;; strategy: function composition 
    ;;          (cases on handle-selected? and range-selected?)
    (define (on-drag mx my)
      (cond
        [handle-selected? (drag-handle mx)]
        [range-selected? (drag-range mx my)]))

    ;; on-button-up : if handle is selected, send command to timer and
    ;; unselect. if range is selected, just unselect
    ;; strategy: function composition
    (define (on-button-up)
      (cond
        [handle-selected? (unselect-handle)]
        [range-selected? (unselect-range)]))

    ;;; helpers for handle operations

    ;; strategy: function composition
    (define (near-handle? mx my)
      (near-rectangle? mx my handle-x-position y HANDLE-WIDTH
        HANDLE-HEIGHT))

    (define (select-handle)
      (set! handle-selected? true))

    (define (drag-handle mx)
      (set! handle-x-position mx))

    (define (unselect-handle)
      (send timer set-time! (handle-x-to-value handle-x-position))
      (set! handle-selected? false))

    ;;; helpers for range operations.
    
    (define (near-range? mx my)
      (near-rectangle? mx my x y RANGE-WIDTH RANGE-HEIGHT))

    ;;  we need some extra state here:

    ;; position of button-down when range was last selected.
    ;; recorded relative to (x,y)
    (field [last-button-down-x 0])
    (field [last-button-down-y 0])

    ;; record the position of the button-down
    (define (select-range mx my)
      (set! last-button-down-x (- mx x))
      (set! last-button-down-y (- my y))
      (set! range-selected? true))

    ;; move center of slider to same position relative to the mouse.
    (define (drag-range mx my)
      ;; jump drag would be:
      ;; (set! x mx)
      ;; (set! y my)
      (set! x (- mx last-button-down-x))
      (set! y (- my last-button-down-y))
      (set! handle-x-position (value-to-handle-x timer-value)))

    (define (unselect-range)
      (set! range-selected? false))

    ;; Integer -> Integer
    ;; Domain Knowledge: Linear interpolation between 0 and MAX-TIMER-VALUE.
    (define (value-to-handle-x v)
      (truncate
        (+ (- x HALF-LENGTH) ;; corresponds to 0
          (* RANGE-WIDTH (/ v MAX-TIMER-VALUE)))))

    ;; Integer -> Integer
    (define (handle-x-to-value hx)
      (truncate
        (* (/ 
             (- hx (- x HALF-LENGTH))   ; distance from left end
             RANGE-WIDTH)           ; as a fraction of the range width
          MAX-TIMER-VALUE)))        ; as a fraction of maximum value

    ;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    ;; on-key : key events are ignored

    (define/public (on-key kev) false)
      
    ))

(define TOLERANCE 5)

;; a point is near a rectangle if it's within TOLERANCE in both x and
;; y.
;; given the coordinates of the point px and py, and the rectangle at
;; center (x,y) and dimensions (w,h)
(define (near-rectangle? px py x y w h)
  (and
    (<= (- x (/ w 2) TOLERANCE) px (+ x (/ w 2) TOLERANCE))
    (<= (- y (/ h 2) TOLERANCE) py (+ y (/ h 2) TOLERANCE))))


        

